from __future__ import annotations

import pickle
from typing import Any, Callable

import cloudpickle


def include_module(name) -> None:
    # cloudpickle.register_pickle_by_value wants an imported module object,
    # but there is really no reason to go through that complication, as
    # it might be prone to errors.
    cloudpickle.cloudpickle._PICKLE_BY_VALUE_MODULES.add(name)


def include_package_from_path(raw_path: str) -> None:
    from pathlib import Path

    path = Path(raw_path).resolve()
    parent = path
    while (parent.parent / "__init__.py").exists():
        parent = parent.parent

    if parent != path:
        include_module(parent.name)


def include_modules_from(obj: Any) -> None:
    module_name = getattr(obj, "__module__", None)
    if not module_name:
        return

    if "." in module_name:
        # Just include the whole package
        package_name, *_ = module_name.partition(".")
        include_module(package_name)
        return

    if module_name == "__main__":
        # When the module is __main__, we need to recursively go up the
        # tree to locate the actual package name.
        import __main__

        include_package_from_path(__main__.__file__)
        return

    include_module(module_name)


def _register(cls: Any, func: Callable) -> None:
    cloudpickle.Pickler.dispatch[cls] = func


def _patch_pydantic_field_serialization() -> None:
    # Cythonized pydantic fields can't be serialized automatically, so we are
    # have a special case handling for them that unpacks it to a dictionary
    # and then reloads it on the other side.
    # https://github.com/ray-project/ray/blob/842bbcf4236e41f58d25058b0482cd05bfe9e4da/python/ray/_private/pydantic_compat.py#L80
    try:
        from pydantic.fields import ModelField, ModelPrivateAttr
    except ImportError:
        return

    def create_model_field(kwargs: dict) -> ModelField:
        return ModelField(**kwargs)

    def pickle_model_field(field: ModelField) -> tuple[Callable, tuple]:
        kwargs = {
            "name": field.name,
            # outer_type_ is the original type for ModelFields,
            # while type_ can be updated later with the nested type
            # like int for List[int].
            "type_": field.outer_type_,
            "class_validators": field.class_validators,
            "model_config": field.model_config,
            "default": field.default,
            "default_factory": field.default_factory,
            "required": field.required,
            "alias": field.alias,
            "field_info": field.field_info,
        }

        return create_model_field, (kwargs,)

    def create_private_attr(kwargs: dict) -> ModelPrivateAttr:
        return ModelPrivateAttr(**kwargs)

    def pickle_private_attr(field: ModelPrivateAttr) -> tuple[Callable, tuple]:
        kwargs = {
            "default": field.default,
            "default_factory": field.default_factory,
        }

        return create_private_attr, (kwargs,)

    _register(ModelField, pickle_model_field)
    _register(ModelPrivateAttr, pickle_private_attr)


def _patch_pydantic_model_serialization() -> None:
    # If user has created new pydantic models in his namespace, we will try to pickle
    # those by value, which means recreating class skeleton, which will stumble upon
    # __pydantic_parent_namespace__ in its __dict__ and it may contain modules that
    # happened to be imported in the namespace but are not actually used, resulting
    # in pickling errors. Unfortunately this also means that `model_rebuid()` might
    # not work.
    try:
        import pydantic
    except ImportError:
        return

    # https://github.com/pydantic/pydantic/pull/2573
    if not hasattr(pydantic, "__version__") or pydantic.__version__.startswith("1."):
        return

    backup = "_original_extract_class_dict"
    if getattr(cloudpickle.cloudpickle, backup, None):
        return

    original = cloudpickle.cloudpickle._extract_class_dict

    def patched(cls):
        attr_name = "__pydantic_parent_namespace__"
        if issubclass(cls, pydantic.BaseModel) and getattr(cls, attr_name, None):
            setattr(cls, attr_name, None)

        return original(cls)

    cloudpickle.cloudpickle._extract_class_dict = patched
    setattr(cloudpickle.cloudpickle, backup, original)


def _patch_lru_cache() -> None:
    # https://github.com/cloudpipe/cloudpickle/issues/178
    # https://github.com/uqfoundation/dill/blob/70f569b0dd268d2b1e85c0f300951b11f53c5d53/dill/_dill.py#L1429

    from functools import _lru_cache_wrapper as LRUCacheType
    from functools import lru_cache

    def create_lru_cache(func: Callable, kwargs: dict) -> LRUCacheType:
        return lru_cache(**kwargs)(func)

    def pickle_lru_cache(obj: LRUCacheType) -> tuple[Callable, tuple]:
        if hasattr(obj, "cache_parameters"):
            params = obj.cache_parameters()
            kwargs = {
                "maxsize": params["maxsize"],
                "typed": params["typed"],
            }
        else:
            kwargs = {"maxsize": obj.cache_info().maxsize}

        return create_lru_cache, (obj.__wrapped__, kwargs)

    _register(LRUCacheType, pickle_lru_cache)


def _patch_lock() -> None:
    # https://github.com/uqfoundation/dill/blob/70f569b0dd268d2b1e85c0f300951b11f53c5d53/dill/_dill.py#L1310
    from _thread import LockType
    from threading import Lock

    def create_lock(locked: bool) -> Lock:
        lock = Lock()
        if locked and not lock.acquire(False):
            raise pickle.UnpicklingError("Cannot acquire lock")
        return lock

    def pickle_lock(obj: LockType) -> tuple[Callable, tuple]:
        return create_lock, (obj.locked(),)

    _register(LockType, pickle_lock)


def _patch_rlock() -> None:
    # https://github.com/uqfoundation/dill/blob/70f569b0dd268d2b1e85c0f300951b11f53c5d53/dill/_dill.py#L1317
    from _thread import RLock as RLockType  # type: ignore[attr-defined]

    def create_rlock(count: int, owner: int) -> RLockType:
        lock = RLockType()
        if owner is not None:
            lock._acquire_restore((count, owner))  # type: ignore[attr-defined]
        if owner and not lock._is_owned():  # type: ignore[attr-defined]
            raise pickle.UnpicklingError("Cannot acquire lock")
        return lock

    def pickle_rlock(obj: RLockType) -> tuple[Callable, tuple]:
        r = obj.__repr__()
        count = int(r.split("count=")[1].split()[0].rstrip(">"))
        owner = int(r.split("owner=")[1].split()[0])

        return create_rlock, (count, owner)

    _register(RLockType, pickle_rlock)


def _patch_console_thread_locals() -> None:
    try:
        from rich.console import ConsoleThreadLocals
    except ModuleNotFoundError:
        return

    def create_locals(kwargs: dict) -> ConsoleThreadLocals:
        return ConsoleThreadLocals(**kwargs)

    def pickle_locals(obj: ConsoleThreadLocals) -> tuple[Callable, tuple]:
        kwargs = {
            "theme_stack": obj.theme_stack,
            "buffer": obj.buffer,
            "buffer_index": obj.buffer_index,
        }
        return create_locals, (kwargs,)

    _register(ConsoleThreadLocals, pickle_locals)


def _patch_exceptions() -> None:
    # Support chained exceptions
    from tblib.pickling_support import install

    install()


def patch_pickle() -> None:
    _patch_pydantic_field_serialization()
    _patch_pydantic_model_serialization()
    _patch_lru_cache()
    _patch_lock()
    _patch_rlock()
    _patch_console_thread_locals()
    _patch_exceptions()

    include_module("fal")



from __future__ import annotations

import os

from fal.config import Config


def bool_envvar(name: str):
    if name in os.environ:
        val = os.environ[name].lower().strip()
        return val != "0" and val != "false" and val != ""
    return False


DEBUG = bool_envvar("DEBUG")
TEST_MODE = bool_envvar("ISOLATE_TEST_MODE")
AUTH_DISABLED = bool_envvar("ISOLATE_AUTH_DISABLED")

config = Config()
config_host = config.get("host")

# FAL_HOST takes precedence over config.host
GRPC_HOST = os.getenv("FAL_HOST") or config_host or "api.alpha.fal.ai"
if not TEST_MODE:
    assert GRPC_HOST.startswith("api"), "FAL_HOST must start with 'api'"

REST_HOST = GRPC_HOST.replace("api", "rest", 1)
REST_SCHEME = "http" if TEST_MODE or AUTH_DISABLED else "https"
REST_URL = f"{REST_SCHEME}://{REST_HOST}"

# fal.run / env.fal.run
FAL_RUN_HOST = (
    GRPC_HOST.replace("api.", "", 1).replace("alpha.", "", 1).replace(".ai", ".run", 1)
)

DONT_OPEN_LINKS = bool_envvar("FAL_DONT_OPEN_LINKS")



from enum import Enum


class LockReason(str, Enum):
    ADMIN_LOCK_PLEASE_CONTACT_HELLOFAL_AI = "Admin lock. Please contact hello@fal.ai."
    EXHAUSTED_BALANCE_TOP_UP_YOUR_BALANCE_AT_FAL_AIDASHBOARDBILLING = (
        "Exhausted balance. Top up your balance at fal.ai/dashboard/billing"
    )
    PLEASE_ADD_A_PAYMENT_METHOD_AT_FAL_AIDASHBOARDBILLING = "Please add a payment method at fal.ai/dashboard/billing"
    THIS_ACCOUNT_HAS_BEEN_ARCHIVED_PLEASE_CONTACT_HELLOFAL_AI = (
        "This account has been archived. Please contact hello@fal.ai."
    )
    UNKNOWN_PLEASE_CONTACT_HELLOFAL_AI = "Unknown. Please contact hello@fal.ai."
    USER_BUDGET_IS_EXCEEDED_ADJUST_IT_AT_FAL_AIDASHBOARDBILLING = (
        "User budget is exceeded. Adjust it at fal.ai/dashboard/billing"
    )

    def __str__(self) -> str:
        return str(self.value)



from __future__ import annotations

from openapi_fal_rest.client import Client

import fal.flags as flags
from fal.sdk import get_default_credentials


class CredentialsClient(Client):
    def get_headers(self) -> dict[str, str]:
        creds = get_default_credentials()
        return {
            **creds.to_headers(),
            **self.headers,
        }


# TODO: accept more auth methods
REST_CLIENT = CredentialsClient(
    flags.REST_URL,
    timeout=30,
    verify_ssl=not flags.TEST_MODE,
    raise_on_unexpected_status=False,
    follow_redirects=True,
)



from fal.auth import current_user_info
from fal.cli import profile
from fal.sdk import get_default_credentials


def _login(args):
    from fal.auth import login
    from fal.console.icons import CHECK_ICON, CROSS_ICON
    from fal.exceptions import FalServerlessException

    try:
        login(args.console)
        args.console.print(f"{CHECK_ICON} Authenticated successfully, welcome!")
    except FalServerlessException as e:
        args.console.print(f"{CROSS_ICON} {e}")
        return

    _unset_account(args)
    _set_account(args)


def _logout(args):
    from fal.auth import logout
    from fal.console.icons import CHECK_ICON, CROSS_ICON
    from fal.exceptions import FalServerlessException

    try:
        logout(args.console)
        args.console.print(f"{CHECK_ICON} Logged out of [cyan bold]fal[/]. Bye!")
    except FalServerlessException as e:
        args.console.print(f"{CROSS_ICON} {e}")
        return

    _unset_account(args)


def _list_accounts(args):
    from rich.style import Style
    from rich.table import Table

    from fal.auth import UserAccess
    from fal.config import Config

    user_access = UserAccess()
    config = Config()
    current_account_name = config.get_internal("team") or user_access.info["nickname"]

    table = Table(border_style=Style(frame=False), show_header=False)
    table.add_column("#")
    table.add_column("Nickname")
    table.add_column("Type")

    for idx, account in enumerate(user_access.accounts):
        selected = account["nickname"] == current_account_name
        color = "bold yellow" if selected else None

        table.add_row(
            f"* {idx + 1}" if selected else f"  {idx + 1}",
            account["nickname"],
            "Personal" if account["is_personal"] else "Team",
            style=color,
        )

    args.console.print(table)


def _unset_account(args):
    from fal.config import Config

    with Config().edit() as config:
        config.unset_internal("team")


def _set_account(args):
    from rich.prompt import Prompt

    from fal.auth import UserAccess
    from fal.config import Config

    user_access = UserAccess()

    if hasattr(args, "account") and args.account:
        if args.account.isdigit():
            acc_index = int(args.account) - 1
            account = user_access.accounts[acc_index]
        else:
            account = user_access.get_account(args.account)
    else:
        _list_accounts(args)
        indices = list(map(str, range(1, len(user_access.accounts) + 1)))
        team_names = [account["nickname"] for account in user_access.accounts]
        acc_choice = Prompt.ask(
            "Select an account by number",
            choices=indices + team_names,
            show_choices=False,
        )
        if acc_choice in indices:
            acc_index = int(acc_choice) - 1
            account = user_access.accounts[acc_index]
        else:
            account = user_access.get_account(acc_choice)

    if account["is_personal"]:
        args.console.print(
            f"Using personal account [cyan]{account['nickname']}[/]. "
            "You can change this later with [bold]fal team set[/]"
        )
    else:
        args.console.print(
            f"Using team account [cyan]{account['nickname']}[/]. "
            "You can change this later with [bold]fal team set[/]"
        )

    with Config().edit() as config:
        config.set_internal("team", account["nickname"])

        # Unset the profile if set
        if current_profile := config.get_internal("profile"):
            args.console.print(
                f"\n[yellow]Unsetting profile [cyan]{current_profile}[/] "
                "to make team selection effective.[/]"
            )
            profile._unset(args, config=config)


def _whoami(args):
    creds = get_default_credentials()
    user_info = current_user_info(creds.to_headers())

    full_name = user_info["full_name"]
    nickname = user_info["nickname"]
    user_id = user_info["user_id"]

    args.console.print(f"Hello, {full_name}: {nickname!r} - {user_id!r}")


def add_parser(main_subparsers, parents):
    auth_help = "Authenticate with fal."
    parser = main_subparsers.add_parser(
        "auth",
        description=auth_help,
        help=auth_help,
        parents=parents,
    )

    subparsers = parser.add_subparsers(
        title="Commands",
        metavar="command",
        dest="cmd",
        required=True,
    )

    login_help = "Log in a user."
    login_parser = subparsers.add_parser(
        "login",
        description=login_help,
        help=login_help,
        parents=parents,
    )
    login_parser.set_defaults(func=_login)

    logout_help = "Log out the currently logged-in user."
    logout_parser = subparsers.add_parser(
        "logout",
        description=logout_help,
        help=logout_help,
        parents=parents,
    )
    logout_parser.set_defaults(func=_logout)

    whoami_help = "Show the currently authenticated user."
    whoami_parser = subparsers.add_parser(
        "whoami",
        description=whoami_help,
        help=whoami_help,
        parents=parents,
    )
    whoami_parser.set_defaults(func=_whoami)



from fal.cli.auth import _list_accounts, _set_account, _unset_account


def add_parser(main_subparsers, parents):
    teams_help = "Manage teams."
    parser = main_subparsers.add_parser(
        "teams",
        aliases=["team"],
        description=teams_help,
        help=teams_help,
        parents=parents,
    )

    subparsers = parser.add_subparsers(
        title="Commands",
        metavar="command",
        dest="cmd",
        required=True,
    )

    list_help = "List teams."
    list_parser = subparsers.add_parser(
        "list",
        description=list_help,
        help=list_help,
        parents=parents,
    )
    list_parser.set_defaults(func=_list_accounts)

    set_help = "Set the current team."
    set_parser = subparsers.add_parser(
        "set",
        description=set_help,
        help=set_help,
        parents=parents,
    )
    set_parser.add_argument("account", help="The team to set.")
    set_parser.set_defaults(func=_set_account)

    unset_help = "Unset the current team."
    unset_parser = subparsers.add_parser(
        "unset",
        description=unset_help,
        help=unset_help,
        parents=parents,
    )
    unset_parser.set_defaults(func=_unset_account)



from typing import Optional, Tuple
from unittest.mock import MagicMock, patch

import pytest

from fal.cli.deploy import _deploy
from fal.cli.main import parse_args
from fal.project import find_project_root


def test_deploy():
    args = parse_args(["deploy", "myfile.py::MyApp"])
    assert args.func == _deploy
    assert args.app_ref == ("myfile.py", "MyApp")


@pytest.fixture
def mock_parse_pyproject_toml():
    return {
        "apps": {
            "my-app": {
                "ref": "src/my_app/inference.py::MyApp",
                "auth": "shared",
                "deployment_strategy": "rolling",
            },
            "another-app": {
                "ref": "src/another_app/inference.py::AnotherApp",
            },
            "app-with-extras": {
                "ref": "src/app_with_extras/inference.py::AppWithExtras",
                "extra_key": "extra_value",
            },
        }
    }


def mock_args(
    app_ref: Tuple[str, Optional[str]],
    app_name: Optional[str] = None,
    auth: Optional[str] = None,
    strategy: Optional[str] = None,
    reset_scale: bool = False,
):
    args = MagicMock()

    args.app_ref = app_ref
    args.app_name = app_name
    args.auth = auth
    args.strategy = strategy
    args.app_scale_settings = reset_scale

    return args


@patch("fal.cli._utils.find_pyproject_toml", return_value="pyproject.toml")
@patch("fal.cli._utils.parse_pyproject_toml")
@patch("fal.cli.deploy._deploy_from_reference")
def test_deploy_with_toml_success(
    mock_deploy_ref, mock_parse_toml, mock_find_toml, mock_parse_pyproject_toml
):
    # Mocking the parse_pyproject_toml function to return a predefined dict
    mock_parse_toml.return_value = mock_parse_pyproject_toml

    args = mock_args(app_ref=("my-app", None))

    _deploy(args)

    project_root, _ = find_project_root(None)

    # Ensure the correct app is deployed
    mock_deploy_ref.assert_called_once_with(
        (f"{project_root / 'src/my_app/inference.py'}", "MyApp"),
        "my-app",
        args,
        "shared",
        "rolling",
        scale=False,
    )


@patch("fal.cli._utils.find_pyproject_toml", return_value="pyproject.toml")
@patch("fal.cli._utils.parse_pyproject_toml")
@patch("fal.cli.deploy._deploy_from_reference")
def test_deploy_with_toml_no_auth(
    mock_deploy_ref, mock_parse_toml, mock_find_toml, mock_parse_pyproject_toml
):
    # Mocking the parse_pyproject_toml function to return a predefined dict
    mock_parse_toml.return_value = mock_parse_pyproject_toml

    args = mock_args(app_ref=("another-app", None))

    _deploy(args)

    project_root, _ = find_project_root(None)

    # Since auth is not provided for "another-app", it should default to "private"
    mock_deploy_ref.assert_called_once_with(
        (f"{project_root / 'src/another_app/inference.py'}", "AnotherApp"),
        "another-app",
        args,
        "private",
        "recreate",
        scale=False,
    )


@patch("fal.cli._utils.find_pyproject_toml", return_value="pyproject.toml")
@patch("fal.cli._utils.parse_pyproject_toml")
@patch("fal.cli.deploy._deploy_from_reference")
def test_deploy_with_toml_app_not_found(
    mock_deploy_ref, mock_parse_toml, mock_find_toml, mock_parse_pyproject_toml
):
    # Mocking the parse_pyproject_toml function to return a predefined dict
    mock_parse_toml.return_value = mock_parse_pyproject_toml

    args = mock_args(app_ref=("non-existent-app", None))

    # Expect a ValueError since "non-existent-app" is not in the toml
    with pytest.raises(
        ValueError, match="App non-existent-app not found in pyproject.toml"
    ):
        _deploy(args)


@patch("fal.cli._utils.find_pyproject_toml", return_value="pyproject.toml")
@patch("fal.cli._utils.parse_pyproject_toml")
@patch("fal.cli.deploy._deploy_from_reference")
def test_deploy_with_toml_missing_ref_key(
    mock_deploy_ref, mock_parse_toml, mock_find_toml
):
    # Mocking a toml structure without a "ref" key for a certain app
    mock_parse_toml.return_value = {
        "apps": {
            "my-app": {
                "auth": "shared",
            }
        }
    }

    args = mock_args(app_ref=("my-app", None))

    # Expect a ValueError since "ref" key is missing
    with pytest.raises(
        ValueError, match="App my-app does not have a ref key in pyproject.toml"
    ):
        _deploy(args)


@patch("fal.cli._utils.find_pyproject_toml", return_value="pyproject.toml")
@patch("fal.cli._utils.parse_pyproject_toml")
@patch("fal.cli.deploy._deploy_from_reference")
def test_deploy_with_toml_extra_keys_in_toml(
    mock_deploy_ref, mock_parse_toml, mock_find_toml, mock_parse_pyproject_toml
):
    mock_parse_toml.return_value = mock_parse_pyproject_toml

    args = mock_args(app_ref=("app-with-extras", None))

    with pytest.raises(
        ValueError,
        match="Found unexpected keys in pyproject.toml: {'extra_key': 'extra_value'}",
    ):
        _deploy(args)


@patch("fal.cli._utils.find_pyproject_toml", return_value=None)
def test_deploy_with_toml_file_not_found(mock_find_toml):
    args = mock_args(app_ref=("my-app", None))

    # Expect a ValueError since no pyproject.toml file is found
    with pytest.raises(ValueError, match="No pyproject.toml file found."):
        _deploy(args)


@patch("fal.cli._utils.find_pyproject_toml", return_value="pyproject.toml")
@patch("fal.cli._utils.parse_pyproject_toml")
def test_deploy_with_toml_only_app_name_is_provided(
    mock_parse_toml, mock_find_toml, mock_parse_pyproject_toml
):
    mock_parse_toml.return_value = mock_parse_pyproject_toml

    args = mock_args(
        app_ref=("my-app", None), app_name="custom-app-name", auth="public"
    )

    # Expect a ValueError since --app-name and --auth cannot be used with just
    # the app name reference
    with pytest.raises(
        ValueError, match="Cannot use --app-name or --auth with app name reference."
    ):
        _deploy(args)


@patch("fal.cli._utils.find_pyproject_toml", return_value="pyproject.toml")
@patch("fal.cli._utils.parse_pyproject_toml")
@patch("fal.cli.deploy._deploy_from_reference")
def test_deploy_with_toml_deployment_strategy(
    mock_deploy_ref, mock_parse_toml, mock_find_toml, mock_parse_pyproject_toml
):
    mock_parse_toml.return_value = mock_parse_pyproject_toml

    args = mock_args(app_ref=("my-app", None), strategy="rolling")

    _deploy(args)

    project_root, _ = find_project_root(None)

    mock_deploy_ref.assert_called_once_with(
        (f"{project_root / 'src/my_app/inference.py'}", "MyApp"),
        "my-app",
        args,
        "shared",
        "rolling",
        scale=False,
    )


@patch("fal.cli._utils.find_pyproject_toml", return_value="pyproject.toml")
@patch("fal.cli._utils.parse_pyproject_toml")
@patch("fal.cli.deploy._deploy_from_reference")
def test_deploy_with_toml_default_deployment_strategy(
    mock_deploy_ref, mock_parse_toml, mock_find_toml, mock_parse_pyproject_toml
):
    mock_parse_toml.return_value = mock_parse_pyproject_toml

    args = mock_args(app_ref=("another-app", None))

    _deploy(args)

    project_root, _ = find_project_root(None)

    mock_deploy_ref.assert_called_once_with(
        (f"{project_root / 'src/another_app/inference.py'}", "AnotherApp"),
        "another-app",
        args,
        "private",
        "recreate",
        scale=False,
    )


@patch("fal.cli._utils.find_pyproject_toml", return_value="pyproject.toml")
@patch("fal.cli._utils.parse_pyproject_toml")
@patch("fal.cli.deploy._deploy_from_reference")
def test_deploy_with_cli_auth(
    mock_deploy_ref, mock_parse_toml, mock_find_toml, mock_parse_pyproject_toml
):
    mock_parse_toml.return_value = mock_parse_pyproject_toml

    args = mock_args(app_ref=("src/my_app/inference.py", "MyApp"), auth="shared")

    _deploy(args)

    mock_deploy_ref.assert_called_once_with(
        ("src/my_app/inference.py", "MyApp"),
        None,
        args,
        "shared",
        None,
        scale=False,
    )


@patch("fal.cli._utils.find_pyproject_toml", return_value="pyproject.toml")
@patch("fal.cli._utils.parse_pyproject_toml")
@patch("fal.cli.deploy._deploy_from_reference")
def test_deploy_with_cli_deployment_strategy(
    mock_deploy_ref, mock_parse_toml, mock_find_toml, mock_parse_pyproject_toml
):
    mock_parse_toml.return_value = mock_parse_pyproject_toml

    args = mock_args(app_ref=("src/my_app/inference.py", "MyApp"), strategy="rolling")

    _deploy(args)

    mock_deploy_ref.assert_called_once_with(
        ("src/my_app/inference.py", "MyApp"),
        None,
        args,
        None,
        "rolling",
        scale=False,
    )


@patch("fal.cli._utils.find_pyproject_toml", return_value="pyproject.toml")
@patch("fal.cli._utils.parse_pyproject_toml")
@patch("fal.cli.deploy._deploy_from_reference")
def test_deploy_with_cli_reset_scale(
    mock_deploy_ref, mock_parse_toml, mock_find_toml, mock_parse_pyproject_toml
):
    mock_parse_toml.return_value = mock_parse_pyproject_toml

    args = mock_args(app_ref=("src/my_app/inference.py", "MyApp"), reset_scale=True)

    _deploy(args)

    mock_deploy_ref.assert_called_once_with(
        ("src/my_app/inference.py", "MyApp"),
        None,
        args,
        None,
        None,
        scale=True,
    )


@patch("fal.cli._utils.find_pyproject_toml", return_value="pyproject.toml")
@patch("fal.cli._utils.parse_pyproject_toml")
@patch("fal.cli.deploy._deploy_from_reference")
def test_deploy_with_cli_scale(
    mock_deploy_ref, mock_parse_toml, mock_find_toml, mock_parse_pyproject_toml
):
    mock_parse_toml.return_value = mock_parse_pyproject_toml

    args = mock_args(app_ref=("src/my_app/inference.py", "MyApp"))

    _deploy(args)

    mock_deploy_ref.assert_called_once_with(
        ("src/my_app/inference.py", "MyApp"),
        None,
        args,
        None,
        None,
        scale=False,
    )



from __future__ import annotations

import tempfile
import unittest.mock
from pathlib import Path
from typing import Callable
from uuid import uuid4

import pytest
from pydantic import BaseModel, Field
from pydantic import __version__ as pydantic_version

import fal
from fal import FalServerlessHost, FalServerlessKeyCredentials, local, sync_dir
from fal.api import FalServerlessError, IsolatedFunction
from fal.toolkit import (
    File,
    clone_repository,
    download_file,
    download_model_weights,
)
from fal.toolkit import (
    Image as FalImage,
)
from fal.toolkit.file.file import CompressedFile
from fal.toolkit.utils.download_utils import _git_rev_parse, _hash_url

EXAMPLE_FILE_URL = "https://raw.githubusercontent.com/fal-ai/fal/main/projects/fal/tests/assets/cat.png"


@pytest.mark.flaky(max_runs=3)
def test_isolated(isolated_client: Callable[..., Callable[..., IsolatedFunction]]):
    @isolated_client("virtualenv", requirements=["pyjokes==0.5.0"])
    def get_pyjokes_version():
        import pyjokes

        return pyjokes.__version__

    result = get_pyjokes_version()
    assert result == "0.5.0"

    @isolated_client("virtualenv")
    def get_hostname() -> str:
        import socket

        hostname = socket.gethostname()
        return hostname

    import socket

    local_hostname = socket.gethostname()

    first = get_hostname()
    assert local_hostname != first

    get_hostname_local = get_hostname.on(local)
    second = get_hostname_local()
    assert local_hostname == second

    get_hostname_m = get_hostname.on(machine_type="L")
    third = get_hostname_m()
    assert local_hostname != third


def test_isolate_setup_funcs(isolated_client):
    def setup_function():
        import math

        return math.pi

    @isolated_client(setup_function=setup_function)
    def is_tau(setup, by_factor) -> str:
        import math

        return setup * by_factor == math.tau

    assert is_tau(2)
    assert not is_tau(by_factor=3)


def test_isolate_setup_func_order(isolated_client):
    def setup_function():
        return "one "

    @isolated_client(setup_function=setup_function)
    def one_and(setup, num) -> str:
        return setup + num

    assert one_and("two") == "one two"
    assert one_and("three") == "one three"


def test_isolate_error_handling(isolated_client):
    requirements = ["pyjokes", "requests"]

    def setup():
        print("hello")

    @isolated_client(requirements=requirements, keep_alive=20, setup=setup)
    def raises_value_error():
        import pyjokes

        return pyjokes.get_joke()

    with pytest.raises(ValueError):
        raises_value_error()

    creds = FalServerlessKeyCredentials(key_id="fake", key_secret="fake")
    host = FalServerlessHost(url="not.there", credentials=creds)

    @isolated_client(host=host)
    def raises_grpc_error():
        return True

    with pytest.raises(FalServerlessError):
        raises_grpc_error()


@pytest.mark.skip(reason="No way to test this yet")
def test_sync(isolated_client):
    import os
    import random
    import string
    import tempfile

    def create_random_file(file_path, size):
        with open(file_path, "wb") as f:
            f.write(os.urandom(size))

    def create_test_directory():
        temp_dir = tempfile.mkdtemp()
        for i in range(3):
            file_name = "".join(random.choices(string.ascii_letters, k=10)) + ".txt"
            file_path = os.path.join(temp_dir, file_name)
            create_random_file(file_path, 1024)
        return temp_dir

    @isolated_client()
    def list_remote_directory(remote_dir):
        import os

        contents = os.listdir(remote_dir)
        return contents

    @isolated_client()
    def remove_remote_directory(remote_dir):
        import shutil

        shutil.rmtree(remote_dir)

    local_dir = create_test_directory()
    # Copy the basename of the local directory to keep it unique
    remote_dir_name = "/data/" + os.path.basename(local_dir)
    remote_dir_ref = sync_dir(local_dir, remote_dir_name)

    remote_contents = list_remote_directory(remote_dir_ref)
    local_contents = os.listdir(local_dir)

    local_contents.sort()
    remote_contents.sort()

    assert (
        local_contents == remote_contents
    ), "Local and remote directory contents do not match"

    print("Test passed: Local and remote directory contents match")

    # Clean the directory after tests finished
    remove_remote_directory(isolated_client, remote_dir_ref)


@pytest.fixture
def mock_fal_persistent_dirs(monkeypatch):
    """Mock fal persistent directories.
    It is NOT enough to mock only `FAL_PERSISTENT_DIR`, since the other two are
    might already be evaluated based on the `FAL_PERSISTENT_DIR` during imports.
    """
    temp_data_dir = Path(f"/tmp/data/fal-tests-{uuid4().hex}")
    temp_repository_dir = temp_data_dir / ".fal" / "repos"
    temp_model_weights_dir = temp_data_dir / ".fal" / "model_weights"

    monkeypatch.setattr(
        fal.toolkit.utils.download_utils,
        "FAL_PERSISTENT_DIR",
        temp_data_dir,
    )

    monkeypatch.setattr(
        fal.toolkit.utils.download_utils,
        "FAL_REPOSITORY_DIR",
        temp_repository_dir,
    )

    monkeypatch.setattr(
        fal.toolkit.utils.download_utils,
        "FAL_MODEL_WEIGHTS_DIR",
        temp_model_weights_dir,
    )


def test_download_file(isolated_client, mock_fal_persistent_dirs):
    from fal.toolkit.utils.download_utils import FAL_PERSISTENT_DIR

    relative_directory = "test"
    output_directory = FAL_PERSISTENT_DIR / relative_directory
    expected_path = output_directory / "cat.png"

    @isolated_client()
    def absolute_path_persistent_dir():
        downloaded_path = download_file(
            EXAMPLE_FILE_URL,
            target_dir=output_directory,
        )

        return downloaded_path

    assert str(expected_path) == str(
        absolute_path_persistent_dir()
    ), f"Path should be the target location sent '{expected_path!r}'"

    @isolated_client()
    def absolute_path_non_persistent_dir():
        downloaded_path = download_file(
            EXAMPLE_FILE_URL,
            target_dir=output_directory,
        )

        return downloaded_path

    assert str(expected_path) == str(
        absolute_path_non_persistent_dir()
    ), f"Path should be the target location sent '{expected_path!r}'"

    @isolated_client()
    def relative_path():
        downloaded_path = download_file(
            EXAMPLE_FILE_URL,
            target_dir=relative_directory,
        )

        return downloaded_path

    assert str(expected_path) == str(
        relative_path()
    ), f"Path should be the target location sent '{expected_path!r}'"

    @isolated_client()
    def test_with_force():
        first_path = download_file(
            EXAMPLE_FILE_URL,
            target_dir=output_directory,
            force=False,
        )
        first_path_stat = first_path.stat()

        second_path = download_file(
            EXAMPLE_FILE_URL,
            target_dir=output_directory,
            force=False,
        )
        second_path_stat = second_path.stat()

        third_path = download_file(
            EXAMPLE_FILE_URL,
            target_dir=output_directory,
            force=True,
        )
        third_path_stat = third_path.stat()

        return (
            first_path,
            first_path_stat,
            second_path,
            second_path_stat,
            third_path,
            third_path_stat,
        )

    (
        first_path,
        first_path_stat,
        second_path,
        second_path_stat,
        third_path,
        third_path_stat,
    ) = test_with_force()

    assert str(expected_path) == str(first_path), "Path should be the target location"
    assert str(expected_path) == str(second_path), "Path should be the target location"
    assert str(expected_path) == str(third_path), "Path should be the target location"

    assert (
        first_path_stat.st_mtime == second_path_stat.st_mtime
    ), "The file should not be redownloaded"

    assert (
        second_path_stat.st_mtime < third_path_stat.st_mtime
    ), "The file should be redownloaded with force=True"


def test_download_model_weights(isolated_client, mock_fal_persistent_dirs):
    from fal.toolkit.utils.download_utils import FAL_MODEL_WEIGHTS_DIR

    print(FAL_MODEL_WEIGHTS_DIR)

    expected_path = FAL_MODEL_WEIGHTS_DIR / _hash_url(EXAMPLE_FILE_URL) / "cat.png"

    @isolated_client()
    def download_weights():
        first_path = download_model_weights(EXAMPLE_FILE_URL, force=False)
        first_path_stat = first_path.stat()

        second_path = download_model_weights(EXAMPLE_FILE_URL, force=False)
        second_path_stat = second_path.stat()

        third_path = download_model_weights(EXAMPLE_FILE_URL, force=True)
        third_path_stat = third_path.stat()

        return (
            first_path,
            first_path_stat,
            second_path,
            second_path_stat,
            third_path,
            third_path_stat,
        )

    (
        first_path,
        first_path_stat,
        second_path,
        second_path_stat,
        third_path,
        third_path_stat,
    ) = download_weights()

    assert str(expected_path) == str(first_path), "Path should be the target location"
    assert str(expected_path) == str(second_path), "Path should be the target location"
    assert str(expected_path) == str(third_path), "Path should be the target location"

    assert (
        first_path_stat.st_mtime == second_path_stat.st_mtime
    ), "The model weights should not be redownloaded"

    assert (
        second_path_stat.st_mtime < third_path_stat.st_mtime
    ), "The model weights should be redownloaded with force=True"


def test_clone_repository(isolated_client, mock_fal_persistent_dirs):
    from fal.toolkit.utils.download_utils import FAL_REPOSITORY_DIR

    # https://github.com/fal-ai/isolate/tree/64b0a89c8391bd2cb3ca23cdeae01779e11aee05
    EXAMPLE_REPO_URL = "https://github.com/fal-ai/isolate.git"
    EXAMPLE_REPO_FIRST_COMMIT = "64b0a89c8391bd2cb3ca23cdeae01779e11aee05"
    EXAMPLE_REPO_SECOND_COMMIT = "34ecbca8cc7b64719d2a5c40dd3272f8d13bc1d2"
    expected_path = FAL_REPOSITORY_DIR / "isolate"
    first_expected_path = (
        FAL_REPOSITORY_DIR / f"isolate-{EXAMPLE_REPO_FIRST_COMMIT[:8]}"
    )
    second_expected_path = (
        FAL_REPOSITORY_DIR / f"isolate-{EXAMPLE_REPO_SECOND_COMMIT[:8]}"
    )

    @isolated_client()
    def clone_without_commit_hash():
        repo_path = clone_repository(EXAMPLE_REPO_URL)

        return repo_path

    repo_path = clone_without_commit_hash()
    assert str(repo_path) == str(expected_path), "Path should be the target location"

    @isolated_client()
    def clone_with_commit_hash():
        first_path = clone_repository(
            EXAMPLE_REPO_URL, commit_hash=EXAMPLE_REPO_FIRST_COMMIT
        )
        first_repo_hash = _git_rev_parse(first_path, "HEAD")

        second_path = clone_repository(
            EXAMPLE_REPO_URL, commit_hash=EXAMPLE_REPO_SECOND_COMMIT
        )

        second_repo_hash = _git_rev_parse(second_path, "HEAD")

        return first_path, first_repo_hash, second_path, second_repo_hash

    (
        first_path,
        first_repo_hash,
        second_path,
        second_repo_hash,
    ) = clone_with_commit_hash()

    assert str(first_expected_path) == str(
        first_path
    ), "Path should be the target location"
    assert str(second_expected_path) == str(
        second_path
    ), "Path should be the target location"

    assert (
        first_repo_hash == EXAMPLE_REPO_FIRST_COMMIT
    ), "The commit hash should be the same"
    assert (
        second_repo_hash == EXAMPLE_REPO_SECOND_COMMIT
    ), "The commit hash should be the same"

    @isolated_client()
    def clone_with_force():
        first_path = clone_repository(
            EXAMPLE_REPO_URL, commit_hash=EXAMPLE_REPO_FIRST_COMMIT, force=False
        )
        first_repo_stat = first_path.stat()

        second_path = clone_repository(
            EXAMPLE_REPO_URL, commit_hash=EXAMPLE_REPO_FIRST_COMMIT, force=False
        )
        second_repo_stat = second_path.stat()

        third_path = clone_repository(
            EXAMPLE_REPO_URL, commit_hash=EXAMPLE_REPO_FIRST_COMMIT, force=True
        )
        third_repo_stat = third_path.stat()

        return (
            first_path,
            first_repo_stat,
            second_path,
            second_repo_stat,
            third_path,
            third_repo_stat,
        )

    (
        first_path,
        first_repo_stat,
        second_path,
        second_repo_stat,
        third_path,
        third_repo_stat,
    ) = clone_with_force()

    assert str(first_expected_path) == str(
        first_path
    ), "Path should be the target location"
    assert str(first_expected_path) == str(
        second_path
    ), "Path should be the target location"
    assert str(first_expected_path) == str(
        third_path
    ), "Path should be the target location"

    assert (
        first_repo_stat.st_mtime == second_repo_stat.st_mtime
    ), "The repository should not be cloned again"

    assert (
        first_repo_stat.st_mtime < third_repo_stat.st_mtime
    ), "The repository should be cloned again with force=True"


def fal_file_downloaded(file: File):
    return file.file_size is not None


def fal_file_url_matches(file: File, url: str):
    return file.url == url


def fal_file_content_matches(file: File, content: str):
    return file.as_bytes().decode() == content


def test_fal_file_from_path(isolated_client):
    @isolated_client(requirements=[f"pydantic=={pydantic_version}", "tomli"])
    def fal_file_from_temp(content: str):
        with tempfile.NamedTemporaryFile() as temp_file:
            file_path = temp_file.name

            with open(file_path, "w") as fp:
                fp.write(content)

            return File.from_path(file_path, repository="in_memory")

    file_content = "file-test"
    file = fal_file_from_temp(file_content)

    assert fal_file_content_matches(file, file_content)


def test_fal_file_from_bytes(isolated_client):
    @isolated_client(requirements=[f"pydantic=={pydantic_version}", "tomli"])
    def fal_file_from_bytes(content: str):
        return File.from_bytes(content.encode(), repository="in_memory")

    file_content = "file-test"
    file = fal_file_from_bytes(file_content)

    assert fal_file_content_matches(file, file_content)


def test_fal_file_save(isolated_client):
    @isolated_client(requirements=[f"pydantic=={pydantic_version}", "tomli"])
    def fal_file_to_local_file(content: str):
        file = File.from_bytes(content.encode(), repository="in_memory")

        with tempfile.NamedTemporaryFile() as temp_file:
            file_name = temp_file.name
            # We need to overwrite the file, since the file is already created
            # by the tempfile.NamedTemporaryFile
            file.save(file_name, overwrite=True)

            with open(file_name) as fp:
                file_content = fp.read()

        return file_content

    file_content = "file-test"
    saved_file_content = fal_file_to_local_file(file_content)

    assert file_content == saved_file_content


@pytest.mark.parametrize(
    "file_url, expected_content",
    [
        (
            EXAMPLE_FILE_URL,
            "projects/fal/cat.png",
        ),
        ("data:text/plain;charset=UTF-8,fal", "fal"),
    ],
)
def test_fal_file_input(isolated_client, file_url: str, expected_content: str):
    class TestInput(BaseModel):
        file: File = Field()

    @isolated_client(requirements=[f"pydantic=={pydantic_version}", "tomli"])
    def init_file_on_fal(input: TestInput) -> File:
        return input.file

    test_input = TestInput(file=file_url)
    file = init_file_on_fal(test_input)

    # File is not downloaded until it is needed
    assert not fal_file_downloaded(file)

    assert fal_file_url_matches(file, file_url)

    # Expect value error if we try to access the file content for input file
    with pytest.raises(ValueError):
        fal_file_content_matches(file, expected_content)


def test_fal_compressed_file(isolated_client):
    class TestInput(BaseModel):
        files: CompressedFile

    @isolated_client(requirements=[f"pydantic=={pydantic_version}", "tomli"])
    def init_compressed_file_on_fal(input: TestInput) -> int:
        extracted_file_paths = [file for file in input.files]
        return extracted_file_paths

    archive_url = "https://storage.googleapis.com/falserverless/sdk_tests/compressed_file_test.zip"
    test_input = TestInput(files=archive_url)

    extracted_file_paths = init_compressed_file_on_fal(test_input)

    assert all(isinstance(file, Path) for file in extracted_file_paths)
    assert len(extracted_file_paths) == 3


def test_fal_cdn(isolated_client):
    @isolated_client(requirements=[f"pydantic=={pydantic_version}", "tomli"])
    def upload_to_fal_cdn() -> FalImage:
        return FalImage.from_bytes(b"0", "jpeg", repository="cdn")

    uploaded_image = upload_to_fal_cdn()

    assert uploaded_image


def test_download_file_with_slash_in_filename():
    from fal.toolkit.utils.download_utils import _get_remote_file_properties

    test_url = "https://example.com/file/with/slash.txt"

    with unittest.mock.patch(
        "fal.toolkit.utils.download_utils.urlopen"
    ) as mock_urlopen:
        # urlopen is a context manager, so we need to mock the __enter__ method
        mock_response = mock_urlopen.return_value.__enter__.return_value
        mock_response.headers.get_filename.return_value = "file/with/slash.txt"
        mock_response.headers.get.return_value = "100"

        file_name, _ = _get_remote_file_properties(test_url)

    assert "/" not in file_name
